/*
 * @license
 * Licensed Materials - Property of IBM
 * 5724-U18
 * Copyright IBM Corporation. 2016,2019
 */

/**
 * For each type of input. Make sure to create/update the following function:
 * - buildResponseField
 * - restorePreviousValue
 */

Polymer({

  is: 'maximo-response-field',

  behaviors: [ BaseComponent, Inspections ],

  listeners: {
    'carouselDelete': '_deleteAttachment'
  },

  /**
   * Fired when response field is answered or updated.
   *
   * @event update-response-field
   */

  /**
   * Fired when attachment is deleted.
   * 
   * @event delete-attachment
   */

  /**
   * Fired when need a refresh from inspection results.
   * 
   * @event fetch-inspection-result
   */

  /**
   * Fired when response field change its status regarding completion
   * 
   * @event status-changed
   */

  /**
   * Fired when a response field becomes invalid
   * 
   * @event invalid-changed
   */

  /**
   * Fired when a value is set into response field using setResponse
   * 
   * @event field-response-set
   */
  properties: {

	/**
     * Type of event used to select option 
     * from single choice response field
     * ORDER MATTERS - CONSTANT SHOULD COME FIRST HERE
     */
    singleOptionEventType: {
    	type: String,
    	value: function() {
    		return 'click';
    	},
    	readOnly: true
    },
    
    item: {
      type: Object,
      value: null,
      observer: '_itemChanged'
    },

    question: {
      type: Object,
      value: null,
      notify: true
    },

    resultrecord: {
      type: Object,
      readOnly: true
    },
    
    /**
     * Store last valid 
     * inspfieldresult
     */
    fieldResponse: {
      type: Object,
      readOnly: true,
      observer: '_fieldResponseChanged'
    },

    /** 
     * Stores list of meters
     */
    meters: {
      type: Array
    },
    
    /**
     * DEPRECATED
     * Stores attachments of current inspection field
     */
    attachments: {
      type: Array,
      value: []
    },

    /**
     * Complete flag for response field
     * ps. it does not mean question complete
     */
    answered: {
      type: Boolean,
      value: false,
      reflectToAttribute: true,
      readOnly: true,
      observer: '_answeredChanged'
    },

    /**
     * Indicate weather the resp field is required or not
     */
    required: {
      type: Boolean,
      value: false,
      reflectToAttribute: true,
      readOnly: true,
      observer: '_showHideReqIndicator'
    },

    /**
     * Indicate weather the resp field is valid or not
     */
    invalid: {
      type: Boolean,
      value: false,
      reflectToAttribute: true,
      readOnly: true,
      observer: '_invalidChanged'
    },

    /**
     * Prevent collection rendering
     */
    isCharacteristicMeter: {
      type: Boolean,
      value: function() {
        return false;
      }
    }
  },
  
  observers: [
	  'itemRequiredChanged(item.required)'
	  ],
	  
  attached: function() {

    if (!this.item || !this.item.fieldtype) {
      return;
    }
  },

  /****** Observers ******/

  /**
   * Observer item
   * Build response field upon item change 
   */
  _itemChanged: function(newv) {

    // Clear response
    this.resetComponent();

    // Extract lightweight insp result
    if (newv.hasOwnProperty('_lightresult')) {
      newv._id=newv._id+'_'+newv._lightresult.resultnum;
      this._setResultrecord(newv._lightresult);
    }

    this.buildResponseField(newv);

  },
  
  /**
   * Sync field required attribute to 
   */
  itemRequiredChanged: function(newV, oldV) {
	  
	  if (this.required !== newV) {
		  this._setRequired(newV);
	  }
  },

  /**
   * Observer required flag
   */
  _showHideReqIndicator: function(newV) {

    if (newV === true) {
      this.$.requiredIndicator.classList.remove('nonrequired');
    } else {
      this.$.requiredIndicator.classList.add('nonrequired');
    }
    const property = 'required';
    this.itemPropertySetter(property, newV);

  },

  /**
   * Observer answered property
   * Does not fire item observer since it is setting a subproperty
   */
  _answeredChanged: function(newV, oldV) {
	  const property = 'answered';
    this.itemPropertySetter(property, newV);
    this.itemPropertySetter(this._completeProperty, newV);
  },
  
  /**
   * Observer answered property
   * Does not fire item observer since it is setting a subproperty
   */
  _fieldResponseChanged: function(newV, oldV) {
	  const property = 'response';
	  this.itemPropertySetter(property, newV);
  },
  
  /**
   * Observer invalid property
   * Does not fire item observer since it is setting a subproperty
   */
  _invalidChanged: function(newV) {
	  const property = 'invalid';
	  this.itemPropertySetter(property, newV);
  },
  
  /****** Reponse Setters ******/
  setResponse: function(response) {
    
    var item = this.item;
    var type = item.fieldtype_maxvalue;

    switch (type) {
      //Text Response
      case 'TR':
        this.setTextResponse(response);
        break;
      //Number response
      case 'SE':
        this.setNumericResponse(response);
        break;
      // Maximo Meter
      case 'MM':
        this.setMeterResponse(response);
        break;
      // File Upload
      case 'FU':
        this.setFileUploadResponse(response);
        break;
      // Single Option
      case 'SO':
        this.setOptionResponse(response);
        break;
      case 'DO':
        this.setDateOption(response);
        break;
      case 'TO':
        this.setTimeOption(response);
        break;
      case 'DT':
        this.setDateTimeOption(response);
        break;
      default:
        // TODO
        break;
    }

  },

  /**
   * Set meter response according to meter type
   * Fire event to persist response
   */
  setMeterResponse: function(response) {

    var item = this.item;
    var meterType = item.metertype_maxvalue;

    if (meterType === 'CHARACTERISTIC') {
        // Radio option
        this.setOptionResponse(response);
    } else {
      //CONTINUOUS or GAUGE
      this.setNumericResponse(response);
    }
  },

  /**
   * Notify host that response is properly set
   */
  notifyResponseSet: function(details) {
    this.fire('field-response-set', details);
  },

  /**
   * Notify host that response is not valid
   */
  notifyInvalidResponse: function(details) {
    this.fire('invalid-response');
  },

  /**
   * Set text response in input
   * Fire event to persist response
   */
  setTextResponse: function(response) {
    var input = this.$.responseContainer.querySelector('maximo-text');
    input.value = response;
    input.fire('change');
    this.notifyResponseSet();
  },

  /**
   * Set numeric response in input
   * Fire event to persist response
   */
  setNumericResponse: function(response) {
    var procResponse = response.replace(/\D+$/g,'');

    // Validate answer
    var floatVal = parseFloat(procResponse);
    if (isNaN(floatVal)) {
      this.notifyInvalidResponse();
      return;
    }

    var input = this.$.responseContainer.querySelector('maximo-text');
    input.value = procResponse;
    input.fire('change');
    this.notifyResponseSet();
  },

  /**
   * Set date response input
   * Fire event to persist response
   */
  setDateOption: function(response) {
    // Input type not supported by audio
    console.warn('Response type not supported in audio guided');
    this.notifyResponseSet();
  },

  /**
   * Set time response input
   * Fire event to persist response
   */
  setTimeOption: function(response) {

    // Input type not supported by audio
    console.warn('Response type not supported in audio guided');
    this.notifyResponseSet();
    return;

    var regExp = /([0-9]|0[0-9]|1[0-9]|2[0-3]):([0-5][0-9])/g;
    var matchList = response.match(regExp);
    if (!matchList || matchList.length !== 1) {
      this.notifyInvalidResponse();
      return;
    }

    var procResponse = matchList[0];
    var input = this.$.responseContainer.querySelector('maximo-datetime');
    input.value = procResponse;
    input.fire('time-changed');
    this.notifyResponseSet();
  },

  /**
   * Set date and time response input
   * Fire event to persist response
   */
  setDateTimeOption: function(response) {
    // Input type not supported by audio
    console.warn('Response type not supported in audio guided');
    this.notifyResponseSet();
  },

  /**
   * Set file upload response input
   * Fire event to persist response
   */
  setFileUploadResponse: function(response) {
    // Input type not supported by audio
    console.warn('Response type not supported in audio guided');
    this.notifyResponseSet();
  },

  /**
   * Set option response in radio group
   * Fire event to persist response
   */
  setOptionResponse: function(response) {
    // Match response with any option
    var radioOpt = this.matchOption(response);
    if (!radioOpt) {
      console.warn('No option matches the response');
      this.notifyInvalidResponse();
      return;
    }
    
    if (!radioOpt.checked) {
	    // Select option
	    radioOpt.checked = true;
	    // Fire event to persist response
	    var event = new Event(this.singleOptionEventType);
	    radioOpt.dispatchEvent(event);
    }

    this.notifyResponseSet({selectedOption: radioOpt.dataset.optionid});
  },

  /**
   * Look for a option that matches 
   * try by value then try by position
   */
  matchOption: function(response) {
    // Get all radio options
    var radioList = this.$.responseContainer.querySelectorAll('input[type=radio]');
    var optionNodeMatch = null;
    var partialOptionMatches = [];
    var optionStr = 'option';

    // Try matching by content
    for (var i = 0; i < radioList.length; i++) {
      var radioValue = radioList[i].value;
      var optionLowerCaseValue = radioValue.toLowerCase();
      if (optionLowerCaseValue === response.toLowerCase()) {
        optionNodeMatch = radioList[i];
        break;
      }

      // Collect all partial matches
      if (response.length 
          && optionLowerCaseValue.indexOf(response.toLowerCase()) >= 0) {
        // Store option
        partialOptionMatches.push(radioList[i]);
      }
    }

    // Try matching partial content
    if (partialOptionMatches.length === 1) {
      optionNodeMatch = partialOptionMatches[0];
    }

    // Try catching option selected along with `option` string
    var optRegex = new RegExp(optionStr + '\\w?\\s(\\d+)','i');
    var match = optRegex.exec(response);
    var optionMatch = !match ? response : match[match.length-1];

    // Try match position
    var optionPosition = Number.parseInt(optionMatch);
    if (!optionNodeMatch && 
        Number.isInteger(optionPosition) &&
        optionPosition <= radioList.length ) {
      
      optionNodeMatch = radioList[optionPosition-1];
    }

    return optionNodeMatch;
  },

  /****** Getters ******/

  /**
   * Extract available options to this response field
   */
  getFieldOptions: function() {
    var optionList = [];

    var item = this.item;
    var type = item.fieldtype_maxvalue;

    switch (type) {
      // Single Option
      case 'SO':
        optionList = this.getSingleOptionOptions(item);
        break;
      default:
        optionList = [];
        break;
    }

    return optionList; 
  },

  /**
   * Find options available for this single option field 
   */
  getSingleOptionOptions: function(itemfield) {

    var options = [];

    // Get all field options
    var fieldOptions = itemfield.inspfieldoption;
    if (fieldOptions && fieldOptions.length) {
      options = fieldOptions.map(function(opt) {
        return opt.description;
      });
    }

    return options;
  },

  /****** Builders ******/
  
  /**
   * Build response field according to type
   */
  buildResponseField: function(inspField) {

    // Clear container
    this.clearContainers();

    var item = inspField;
    var type = item.fieldtype_maxvalue;
    switch (type) {
      //Text Response
      case 'TR':
        this.buildTextInput();
        break;
      //Number response
      case 'SE':
        this.buildNumberInput();
        break;
      // Maximo Meter
      case 'MM':
        this.buildMeterInput();
        break;
      // File Upload
      case 'FU':
        this.buildAttachmentsInput();
        break;
      // Single Option
      case 'SO':
        this.buildSingleOption();
        break;
      case 'MO':
        this.buildUnavailableInputPreview($M.localize('uitext', 'mxapiinspection','multiple_choice'));
          break;  
      case 'DO':
          this.buildDateTimeOption();
        break;
      case 'TO':
          this.buildDateTimeOption();
        break;
      case 'DT':
          this.buildDateAndTimeOption();
        break;
      // Signature
      case 'SIG':
        this.buildUnavailableInputPreview($M.localize('uitext', 'mxapiinspection','Signature'));
        break;
      default:
        this.buildGenericInput();
        break;
    }
  },

  /**
	 * get decimal separator
	 */
  _getDecimalSeparator:function(){
		 
    var result = '.';
    try {
      let decimalNumber = 1.1;
      let decimalString = decimalNumber.toLocaleString();
      if ($M.getUserInfo() && $M.getUserInfo().defaultLcale){
        var locale = $M.getUserInfo().defaultLcale;
        decimalString = (locale.includes("_")) ? decimalNumber.toLocaleString(locale.split("_")[0]) : decimalNumber.toLocaleString(locale);
      }
      if(decimalString.indexOf('.') != -1){
        result =  '.';
      }else{
        result =  ',';
      }
    } catch(err) {
      console.log('Could not get the decimal separator using the locale.');
    }
    return result;
  },
  /**
   * Build number response input
   */
  buildNumberInput: function() {
    // Set base vars
    var self = this;
    var decimalSeparator = this._getDecimalSeparator();
    var resultRecord = this.resultrecord;
    var itemfield = this.item;

    // Get field type
    var responseFieldType = this.getResponseFieldType(
        itemfield.fieldtype_maxvalue, itemfield.metertype_maxvalue);

    // Create component attributes
    var attrs = {};
    attrs.id = itemfield._id + 'numberresponse';
    attrs.placeholder = $M.localize('uitext', 'mxapiinspresult','numeric_response');
    attrs.type = (decimalSeparator === '.') ?  'number' : 'text';
    attrs.label = itemfield.description;
    
    // Set readonly
    attrs.readonly = false;
    if (!resultRecord || resultRecord.status_maxvalue === 'COMPLETED') {
      attrs.readonly = true;
    }
    
    // Set require flag
    if (itemfield.required === true) {
      this._setRequired(true);
    } else {
      this._setRequired(false);
    }

    // Create component
    var responseField = Polymer.Base.create('maximo-text', attrs);
    
    // Add attribute step = 'any' into the input inside maximo-text. Defect 310674
    var inputNode = responseField.querySelector('input');
    inputNode.setAttribute('step','any');

    // Fetching response 
    //var response = this.filterResult(resultRecord, itemfield);
    var response = itemfield._outerinspfieldresult;

    // Add change listener on input to restore last value case invalid input
    responseField.addEventListener('input', function(event) {
      
      if(decimalSeparator === ',') {
        this.value = this.value.replace(/[^0-9\,]/, '');
      }

      if (!this.$.input.checkValidity()) {
        self.restorePreviousValue();
      }

    });

    // Add change listener on input change
    responseField.addEventListener('change', function(event) {
      event.stopPropagation();
      self.clearMessages();

      var val = (this.value && decimalSeparator === '.') ? Number(this.value) : this.value;
      
      var lastResponse = self.fieldResponse ? self.fieldResponse : response;

      var payload = {
          inspresult: resultRecord,
          respfield: self.prepareResponseField(resultRecord, itemfield, val, lastResponse),
          successCallback: self.updateResponseSuccessHandler.bind(self),
          errorCallback: self.updateResponseFailureHandler.bind(self)
      };

      self.fire('update-response-field', payload);
    });

    // Prevent scroll to top
    responseField.$.label.addEventListener('click', this.preventScroll);

    // Append input in document
    this.$.responseContainer.appendChild(responseField);
    // Set proper ID
    if ('_setIdentification' in responseField) {
      responseField._setIdentification();
    }
    
    // Need to be after inserted in DOM tree
    // Set value and store it locally    
    if (response) {
      if (response[responseFieldType] !== undefined) {
        var respVal = (decimalSeparator === '.') ? response[responseFieldType] : response[responseFieldType + '_localized'];
        responseField.value = respVal;

        // Unecessary since completion has been processed before in container
        // Store response object
        this.async(function() {
          this._responseChanged(responseFieldType, response);
        });
      }
    }

    // Add style
    responseField.setAttribute('style', 'text-align:left;');
    responseField.$.input.setAttribute('style', 'font-size:129%;width:50%;');
    responseField.$.label.setAttribute('style', 'font-size:129%;');
    if (!responseField.label) {
      responseField.$.label.setAttribute('hidden', true);
    }
  },


  /**
   * Build Date Time response input
   */
  buildDateTimeOption: function() {
    // Set base vars
    var self = this;
    var resultRecord = this.resultrecord;
    var itemfield = this.item;
    var response='';

    // Get field type
    var responseFieldType = this.getResponseFieldType(
        itemfield.fieldtype_maxvalue, itemfield.metertype_maxvalue);

    // Create component attributes
    var attrs = {};
    var labelattr={};
   
    labelattr.label = itemfield.description;
    
    attrs.editableType = 'maximo-datetime';
    attrs.label = '';
    
    if(itemfield.fieldtype_maxvalue==='DO'){
    	labelattr.id = itemfield._id + 'datelabel';
    	attrs.id = itemfield._id + 'dateresponse';
    	attrs.type = 'date';
        attrs.editableAttributes = {'type':'date'};
        
    	}
    if(itemfield.fieldtype_maxvalue==='TO'){
    	labelattr.id = itemfield._id + 'timelabel';
    	attrs.id = itemfield._id + 'timeresponse';
    	attrs.type = 'time';
    	attrs.editableAttributes = {'type':'time'};
    	}
    
    attrs.readonly = false;
    
    attrs.editable=true;
    
    if (!resultRecord || resultRecord.status_maxvalue === 'COMPLETED') {
      attrs.readonly = true;
      if(resultRecord && resultRecord.status_maxvalue === 'COMPLETED'){
    	  attrs.editable=false;
      }
    }
    
    // Set require flag
    if (itemfield.required === true) {
      this._setRequired(true);
    } else {
      this._setRequired(false);
    }
    var responseLabel = Polymer.Base.create('maximo-label', labelattr);

    if (attrs.editable){
        var responseField = Polymer.Base.create('maximo-datetime', attrs);
    } else {
        var responseField = Polymer.Base.create('maximo-field', attrs);    
    };
    
    responseField.querySelector("maximo-label").hidden = true;
    
    const formatter = this.formatter();
    // Fetching response 
    if(itemfield.fieldtype_maxvalue==='DO'){

    	response = itemfield._outerinspfieldresult;
    	if (!response){
    		response = "";
    	};
    	// Add change listener on input
    	responseField.addEventListener('date-changed', function(event) {
    		var dateString = formatter.outputDateFormatter(this.value);
    		self.updateResponse(self,dateString,event,resultRecord,itemfield,response);
    	});

	    // Prevent scroll to top
	    responseLabel.$.label.addEventListener('click', this.preventScroll);
	
	    //Append input in document
	    this.$.responseContainer.appendChild(responseLabel);
	    this.$.responseContainer.appendChild(responseField);
	
	    // Set proper ID
	    if ('_setIdentification' in responseField) {
	      responseField._setIdentification();
	    }
	    
	    self.getResponse(itemfield,response,responseFieldType,responseField);    
    }
    if(itemfield.fieldtype_maxvalue==='TO'){
    	response = itemfield._outerinspfieldresult;
    	//Add change listener on input
    	responseField.addEventListener('time-changed', function(event) {
    	      var timeString = formatter.outputTimeFormatter(this.value);
    	      self.updateResponse(self,timeString,event,resultRecord,itemfield,response);
    	});
		//Append input in document
	    this.$.responseContainer.appendChild(responseLabel);
	    this.$.responseContainer.appendChild(responseField);
	
	    // Set proper ID
	    if ('_setIdentification' in responseField) {
	      responseField._setIdentification();
	    }
	
	    self.getResponse(itemfield,response,responseFieldType,responseField);
	}
     
    responseLabel.setAttribute('style', 'text-align:left; width: auto; display:inline-block;');
    responseLabel.$.label.setAttribute('style', 'font-size:129%;');
    if (!responseLabel.label) {
    	responseLabel.$.label.setAttribute('hidden', true);
    }
    
    // Add style
    responseField.setAttribute('style', 'text-align:left; width: auto;display:flex;');
  },
  
  /**
   * Update Response Field Data
   */
  updateResponse:function(self,val,event,resultRecord,itemfield,response,datetimeField){
	  event.stopPropagation();
      self.clearMessages();
      
      var lastResponse = self.fieldResponse ? self.fieldResponse : response;

      var payload = {
          inspresult: resultRecord,
          respfield: self.prepareResponseField(resultRecord, itemfield, val, lastResponse,datetimeField),
          successCallback: self.updateResponseSuccessHandler.bind(self),
          errorCallback: self.updateResponseFailureHandler.bind(self)
      };
      payload.datetimeField=datetimeField;
      self.fire('update-response-field', payload);
  },
  
  /**
   * Get Response Field Value
   */
  getResponse:function(itemfield,response,responseFieldType,responseField){
	  var respVal='';
	  var formatter = this.formatter();

	  if (response && response[responseFieldType] !== undefined) {
		  
		  if(itemfield.fieldtype_maxvalue==='DO' || responseField.type==='date'){
			  respVal =  formatter.parseServerDate(response[responseFieldType]);
		  }
		  
		  if(itemfield.fieldtype_maxvalue==='TO' || responseField.type==='time'){
			  // Priority to localized timeresponse
			  // Localized data directly impacts mobile date and time format
//			  const localizedResponseFieldType = `${responseFieldType}_localized`;
//			  if (localizedResponseFieldType in response) {
//				  respVal = response[localizedResponseFieldType];
//			  } else {
				  respVal = formatter.parseServerTime(response[responseFieldType]);
//			  }
       	  }
		  
		  responseField.value = respVal;
		  // Store response object
		  this.async(function() {
			  this._responseChanged(responseFieldType, response);
		  });
		  
	  }else{
		  
		  responseField.value = respVal;
	  }
  },
  
  /**
   * Formatter for DT, DO and TO types of input
   * 
   * Reference used setDefaultDateTimeFormat(assignment.html) 
   * 
   * @returns {Object}
   */
  formatter: function() {
	  const	profile = $M.maxauth.whoami; // $M.getMaxauth().whoami;
	  const ISODateFormat = 'YYYY-MM-DD';
	  const mobileDateFormat = 'YYYY-MM-DD';
	  const mobileTimeFormat = 'HH:mm';
	  
	  /**
	   * Changing UTC may impact date and time already saved.
	   */
	  let timeResponseUTC = false;
	  
	  return {
		  
		  /**
		   * Used to format date from server
		   * parseZone used to not affect date regression based on timezone change
		   * @param {string} value - date string
		   * @returns {string} datetime formatted string 
		   */
		  parseServerDate: function(value) {
			  let format = this._getDateInputFormat();
			  return moment.parseZone(value).format(format.toUpperCase());
		  },
		  
		  /**
		   * Used to format timeresponse from server
		   * @function
		   * @param {string} value - time string 
		   * @returns {string} datetime formatted string
		   */
		  parseServerTime: function(value) {
			  
			  let format = this._getTimeInputFormat();
			  /**
			   * This can be used to remove offset from datetime
			   */
			  const noZone = this._removeTimezoneOffset(value);
			  /** 
			   * Moment object created with server timezone offset
			   * When timeResponseUTC converts to UTC(Z) otherwise converts to local time 
			   */
			  var formatted = timeResponseUTC ? moment(noZone).utc().format(format) : moment(noZone).format(format);
			  return formatted;
		  },
		  
		  
		  /**
		   * Prepares string to send dateresponse as payload
		   * @function
		   * @param {string} - date string from datepicker
		   * @returns {string}
		   */
		  outputDateFormatter: function(dateString) {
			  var momentObj = this._parseInputDate(dateString);
			  var formattedDate = dateString === "" ? dateString : this._formatOutputDate(momentObj);
			  return formattedDate;
		  },
		  
		  /**
		   * Used to format Date from datetime picker
		   * @function
		   * @param {string} value - date string 
		   * @returns {Moment} 
		   */
		  _parseInputDate: function(value) {
			  var format = this._getDateInputFormat();
			  return moment(value, format.toUpperCase());
		  },
		  
		  /**
		   * Formats moment object to date
		   * @param {Moment} mObject - moment object to be formatted as date
		   * @returns {string}
		   */
		  _formatOutputDate: function(mObject) {
			  let dateFormat = profile.dateformat ? profile.dateformat : ISODateFormat;
			  return mObject.format(dateFormat.toUpperCase());
		  },
		  
		  /**
		   * Prepares string to send timeresponse as payload
		   * @function
		   * @param {string} - time string from datepicker
		   * @returns {string}
		   */
		  outputTimeFormatter: function(timeString) {
			  var momentObj = this._parseInputTime(timeString);
			  var formattedTime = this._formatOutputTime(momentObj);
			  return formattedTime;
		  },

		  /**
		   * Used to format Time from datetime picker
		   * @function
		   * @param {string} value - time string 
		   * @returns {string} 
		   */
		  _parseInputTime: function(value) {
			  var format = this._getTimeInputFormat();
			  // When timeResponseUTC it creates moment object as UTC(Z) otherwise local time
			  var momentObj = timeResponseUTC ? moment.parseZone(value, format) : moment(value, format);
			  return momentObj;
		  },
		  
		  /**
		   * Formats moment object to send timeresponse as payload
		   * @param {Moment} mObject - moment object to be formatted as date
		   * @returns {string}
		   */
		  _formatOutputTime: function(mObject) {
			  // Set first day of the year 
			  mObject.dayOfYear(1);
			  let timeFormat = profile.timeformat ? profile.timeformat : 'HH:mm a';
			  return mObject.format(timeFormat)
		  },
		  
		  /**
		   * Gets correct date format based on user profile and device
		   * @returns {string} date format
		   */
		  _getDateInputFormat: function() {
			  return $M.isMobile() ? mobileDateFormat : profile.dateformat ? profile.dateformat : ISODateFormat;
		  },
		  
		  /**
		   * Gets correct time format based on user profile and device
		   * @returns {string} time format
 		   */
		  _getTimeInputFormat: function() {
			  return $M.isMobile() ? mobileTimeFormat : profile.timeformat ? profile.timeformat : 'HH:mm a';
		  },
		  
		  /**
		   * Removes zone/offset from date and time string
		   * e.g. 1970-01-01T08:49:00-05:00 becomes 1970-01-01T08:49:00
		   * @param {string} dateTimeString - date time format
		   * @returns {string}
		   */
		  _removeTimezoneOffset: function(dateTimeString) {
			  const re = /(?:Z|[+-][01]\d:[0-5]\d)$/;
			  return dateTimeString.replace(re, "");
		  }
		  
	  }
	  
  },
  
  /**
   * Build Date And Time response input
   */
  buildDateAndTimeOption: function() {
	    var self = this;
	    var resultRecord = this.resultrecord;
	    var itemfield = this.item;
	    var responseFieldType='';
	    var responseField='';
	    var response='';
	    
	    // Create component attributes
	    var attrs = {};
	    var labelattr={};
	    labelattr.id = itemfield._id + 'datetimelabel';
	    labelattr.label = itemfield.description;
	    attrs.id = itemfield._id + 'dataandtimeresponse';	  	
	  	attrs.editableType = 'maximo-datetime';  

	    attrs.readonly = false;
	    
	    attrs.editable=true;
	      
	    if (!resultRecord || resultRecord.status_maxvalue === 'COMPLETED') {
	      attrs.readonly = true;
	      if(resultRecord && resultRecord.status_maxvalue === 'COMPLETED'){
		    	attrs.editable=false;
		    }
	    }
	    
	    // Set require flag
	    if (itemfield.required === true) {
	      this._setRequired(true);
	    } else {
	      this._setRequired(false);
	    }
	        
	    var responseLabel = Polymer.Base.create('maximo-label', labelattr);

        // Prevent scroll to top
        responseLabel.$.label.addEventListener('click', this.preventScroll);

	    this.$.responseContainer.appendChild(responseLabel);
	    responseLabel.setAttribute('style', 'text-align:left; width: auto;display:flex;');
	    responseLabel.$.label.setAttribute('style', 'font-size:129%;');
	    if (!responseLabel.label) {
	    	responseLabel.$.label.setAttribute('hidden', true);
	    }
	    
	    const formatter = this.formatter();
	      
	    for(var i=0;i<2;i++){
	    	if(i===0) { 
	      	attrs.id = itemfield._id + 'dateresponse';
	    	attrs.type='date';
	    	attrs.editableAttributes = {'type':'date'};
	    	
	    	if (attrs.editable){
                var responseField = Polymer.Base.create('maximo-datetime', attrs);
    	    } else {
                var responseField = Polymer.Base.create('maximo-field', attrs);    
    	    };
    	    responseField.querySelector("maximo-label").hidden = true;
	    	
	    	responseFieldType = this.getResponseFieldType(
	    			  itemfield.fieldtype_maxvalue, itemfield.metertype_maxvalue,'dateResponse');
	 	    // Fetching response 
	 	    response = itemfield._outerinspfieldresult;
	 	    if(response){
	 	    	response.fieldtype_maxvalue=itemfield.fieldtype_maxvalue;
	 	    	}
	 	   
	 	    // Add change listener on input
	 	    responseField.addEventListener('date-changed', function(event) {
	 	    	var dateString = formatter.outputDateFormatter(this.value);
	 	    	self.updateResponse(self,dateString,event,resultRecord,itemfield,response,'dateResponse');	
	 	    });

	 	    // Append input in document
	        this.$.responseContainer.appendChild(responseField);
	        
	        // Set proper ID
	        if ('_setIdentification' in responseField) {
	          responseField._setIdentification();
	        }
	 	    
	 	   self.getResponse(itemfield,response,responseFieldType,responseField);	 	   
	 	    responseField.setAttribute('style', 'display:inline-block;');
	 	    
    	}
	    	
	    	if(i===1){
	    	attrs.label = '';
	    	attrs.id = itemfield._id + 'timeresponse';
	  	    attrs.type='time';
	    	attrs.editableAttributes = {'type':'time'};	    
	  	    attrs.editableType = 'maximo-datetime';

		    if (attrs.editable){
               var responseField = Polymer.Base.create('maximo-datetime', attrs);
     	    } else {
               var responseField = Polymer.Base.create('maximo-field', attrs);    
    	    };
		    responseField.querySelector("maximo-label").hidden = true;
		    
		    // Fetching response 
		     response = itemfield._outerinspfieldresult;
		     if(response){
		 	    response.fieldtype_maxvalue=itemfield.fieldtype_maxvalue;
		 	 }

		     responseFieldType = this.getResponseFieldType(
		    		 itemfield.fieldtype_maxvalue, itemfield.metertype_maxvalue,'timeResponse');
		    // Add change listener on input
		    responseField.addEventListener('time-changed', function(event) {
		    	var timeString = formatter.outputTimeFormatter(this.value);
		    	self.updateResponse(self,timeString,event,resultRecord,itemfield,response,'timeResponse');
		    });
		   
		    // Append input in document
        this.$.responseContainer.appendChild(responseField);
        
        // Set proper ID
        if ('_setIdentification' in responseField) {
          responseField._setIdentification();
        }

		    self.getResponse(itemfield,response,responseFieldType,responseField);
		    }
	    }		    
	 },

  /**
   * Build text response input
   */
  buildTextInput: function() {
    
    var self = this;
    var resultRecord = this.resultrecord;
    var itemfield = this.item;

    var responseFieldType = this.getResponseFieldType(
        itemfield.fieldtype_maxvalue, itemfield.metertype_maxvalue);

    // Create component attributes
    var attrs = {};
    attrs.id = itemfield._id + 'textresponse';
    attrs.placeholder = $M.localize('uitext', 'mxapiinspection','enter_answer');
    attrs.type = 'text';
    attrs.maxlength = '250';
    attrs.label = itemfield.description;
    
    // Set required
    attrs.readonly = false;
    if (!resultRecord || resultRecord.status_maxvalue === 'COMPLETED') {
      attrs.readonly = true;
    }
    
    // Set require flag
    if (itemfield.required === true) {
      this._setRequired(true);
    } else {
      this._setRequired(false);
    }

    // Fetching response 
    //var response = this.filterResult(resultRecord, itemfield);
    var response = itemfield._outerinspfieldresult;
    
    // Create component
    if( attrs.readonly === true && response && response.txtresponse && response.txtresponse.length > 21 )
    {
    	var responseField = Polymer.Base.create('maximo-textarea', attrs);
    }
    else
    {
    	var responseField = Polymer.Base.create('maximo-text', attrs);
    }

    // Add change listener on input
    responseField.addEventListener('change', function(event) {
      event.stopPropagation();
      self.clearMessages();
      
      var lastResponse = self.fieldResponse ? self.fieldResponse : response;

      var payload = {
          inspresult: resultRecord,
          respfield: self.prepareResponseField(resultRecord, itemfield, this.value, lastResponse),
          successCallback: self.updateResponseSuccessHandler.bind(self),
          errorCallback: self.updateResponseFailureHandler.bind(self)
      };

      self.fire('update-response-field', payload);
    });

    // Prevent scroll to top
    responseField.$.label.addEventListener('click', this.preventScroll);

    // Append input in document
    this.$.responseContainer.appendChild(responseField);
    // Set proper ID
    if ('_setIdentification' in responseField) {
      responseField._setIdentification();
    }
    
    // Need to be after inserted in DOM tree
    // Set value and store it locally    
    if (response) {
      // Store response object
      if (response[responseFieldType] !== undefined) {
        var respVal = response[responseFieldType];
        responseField.value = respVal;
        // Unecessary since completion has been processed before in container
        this.async(function() {
          this._responseChanged(responseFieldType, response);
        });
      }
    }

    // Add style
    responseField.setAttribute('style', 'text-align:left;');
    if(responseField.localName === 'maximo-textarea')
    {
    	responseField.$.input.setAttribute('style', 'font-size:129%;width:100%;');
    }
    else
    {
    	responseField.$.input.setAttribute('style', 'font-size:129%;width:50%;');	
    }
    responseField.$.label.setAttribute('style', 'font-size:129%;');
    if (!responseField.label) {
      responseField.$.label.setAttribute('hidden', true);
    }

  },

  /****** Single Option ******/

  /**
   * Create object to match object supported in mx-radiobutton
   */
  buildOptionObject: function(option) {
    return {
      description: option.description,
      boolVal: false
    };
  },

  /**
   * Build single option component
   */
  buildSingleOption: function() {

    var self = this;
    var resultRecord = this.resultrecord;
    var itemfield = this.item;
    var options = itemfield.inspfieldoption;
    var column1 = Math.ceil(options.length / 2);

    var lastResponse = '';
    var OptionMsgDiv = '';
    var OptionConfirmDiv = '';
    var optionWrapper = '';
    var optionWrapperCol2 = '';

    // Set require flag
    if (itemfield.required === true) {
      this._setRequired(true);
    } else {
      this._setRequired(false);
    }

    var responseFieldType = this.getResponseFieldType(
      itemfield.fieldtype_maxvalue, itemfield.metertype_maxvalue);

    var response = itemfield._outerinspfieldresult;


    var respVal;
    if (response) {
      // Store response object
      if (response[responseFieldType] !== undefined) {
        respVal = response[responseFieldType];
        // Unecessary since completion has been processed before in container
        this.async(function() {
          this._responseChanged(responseFieldType, response);
        });
      }
    }

    // Order by sequence created
    options.sort(function(a, b) {
      return parseFloat(a.sequence) - parseFloat(b.sequence);
    });

    // Create component attributes
    var labelAttrs = {};
    labelAttrs.id = itemfield._id + '_attachmentLabel';
    labelAttrs.label = itemfield.description;

    // Create component
    var label = Polymer.Base.create('maximo-label', labelAttrs);
    var wrapper = Polymer.Base.create('div', {name: 'optionsContainer'});
    var wrapper1 = Polymer.Base.create('div');
    wrapper1.id = 'wrapper1';
    var wrapper2 = Polymer.Base.create('div');
    wrapper2.id = 'wrapper2';
    var OptionMainDiv = Polymer.Base.create('div', {
      'id': 'OptionMainDiv_' + itemfield._id
    });

    var optionSelectionFeedbackContainer = Polymer.Base.create('div', {name: 'optionFeedbackContainer'});

    OptionMainDiv.setAttribute('style', 'display: flex; justify-content: space-between; width:100%; bottom: 0; max-height: 40px; overflow-y: auto; margin: 10px 0;');
    // Prevent scroll to top
    label.$.label.addEventListener('click', this.preventScroll);

    // Append to container
    this.$.responseContainer.appendChild(label);
    var counter = 0;
    // Iterate over response field option
    options.forEach(function(option) {

      var radioBtn = document.createElement('input');
      var radioId = option._id + '_' + itemfield._id + 'singleOptionResponse';

      radioBtn.name = 'radiogroup' + itemfield._id;
      radioBtn.type = 'radio';
      radioBtn.value = option.description;
      radioBtn.id = radioId;
      radioBtn.dataset.optionid = option.inspfieldoptionid;

      radioBtn.icon = option.icon;
      radioBtn.color = option.color;

      if (respVal && respVal === radioBtn.value) {
        radioBtn.checked = true;
      }

      if (!resultRecord || resultRecord.status_maxvalue === status_.COMPLETED) {
        radioBtn.disabled = true;
      }

      radioBtn.addEventListener(this.singleOptionEventType, function(event) {

        event.stopPropagation();
        self.clearMessages();

        for (var i = 0; i < wrapper1.childElementCount; i++) {
          wrapper1.children[i].querySelector('.divOption').style.backgroundColor = self.getComputedStyleValue('--Primary-white');
        }

        for (i = 0; i < wrapper2.childElementCount; i++) {
          wrapper2.children[i].querySelector('.divOption').style.backgroundColor = self.getComputedStyleValue('--Primary-white');
        }

        lastResponse = self.fieldResponse ? self.fieldResponse : response;

        var tempRadioValue = this.value;

        if ((lastResponse !== undefined) && (tempRadioValue === lastResponse.txtresponse)) {
          tempRadioValue = '';
          for (var i2 = 0; i2 < options.length; i2++) {
            for (var i = 0; i < wrapper1.childElementCount; i++) {
              if (wrapper1.children[i].querySelector('input').dataset.optionid == options[i2]._id) {
                wrapper1.children[i].querySelector('.divOption').style.backgroundColor = options[i2].color;
              }
            }

            for (var i = 0; i < wrapper2.childElementCount; i++) {
              if (wrapper2.children[i].querySelector('input').dataset.optionid == options[i2]._id) {
                wrapper2.children[i].querySelector('.divOption').style.backgroundColor = options[i2].color;
              }
            }
          }
        } else {
          if (radioBtn.color && !(radioBtn.color === "") && !(radioBtn.color.toUpperCase() === self.getComputedStyleValue('--Primary-white').toUpperCase())) {
            event.target.closest('.divOption').style.backgroundColor = radioBtn.color;
            event.target.closest('.divOption').style.borderColor = radioBtn.color;
          } else {
            event.target.closest('.divOption').style.backgroundColor = self.getComputedStyleValue('--Primary-green10');
            event.target.closest('.divOption').style.borderColor = self.getComputedStyleValue('--Secondary-blue20');
          }

        };

        resultRecord.event = "change";
        var payload = {
          inspresult: resultRecord,
          respfield: self.prepareResponseField(resultRecord, itemfield, tempRadioValue + '~' + false, lastResponse),
          successCallback: self.updateResponseSuccessHandler.bind(self),
          errorCallback: self.updateResponseFailureHandler.bind(self)
        };

        self.fire('update-response-field', payload);

        wrapper = this.closest('maximo-response-field');

        if (tempRadioValue) {
          wrapper._setInspFeedback(OptionMsgDiv, itemfield,
            OptionConfirmDiv, OptionMainDiv, optionSelectionFeedbackContainer, option, self, resultRecord);
        } else {
          wrapper._setInspFeedback(OptionMsgDiv, itemfield,
            OptionConfirmDiv, OptionMainDiv, optionSelectionFeedbackContainer, {}, self, resultRecord);
        }


      }); // End - eventListener 

      var radioLabel = document.createElement('label');
      radioLabel.setAttribute('for', radioId);
      radioLabel.setAttribute('id', radioId + '_label');
      radioLabel.textContent = option.description;

      // Accessibility aspect for single option
      radioLabel.setAttribute('role','radio');
      radioLabel.setAttribute('tabindex', '0');
      radioLabel.addEventListener('keydown', function(event) {
        switch (event.which){
          case 13: // enter
          case 32: // space
            event.preventDefault();
            var radioInput = this.querySelector('input');
            if (radioInput) radioInput.click();
            break;
        }
      });

      var col = '';
      if (options.length <= 4) {
        col = '1';
      }
      else {
        if (counter < column1) {
          col = '1';
        } else {
          col = '2';
        }
      }

      var status;
      if (resultRecord) {
        if (resultRecord.status_maxvalue === status_.COMPLETED) {
          status = status_.COMPLETED;
        }
      }

      wrapper = this._setSingleOptionStyle(respVal, optionWrapper, optionWrapperCol2, radioBtn, radioLabel, col, wrapper1, wrapper2, wrapper, status);

      counter++;

      if (radioBtn.checked) {
        optionSelectionFeedbackContainer = this._setInspFeedback(OptionMsgDiv, itemfield, OptionConfirmDiv, OptionMainDiv, optionSelectionFeedbackContainer, option, self, resultRecord);
      }

    }, this);

    if (resultRecord && (resultRecord.status_maxvalue === status_.INPROGRESS ||
      resultRecord.status_maxvalue === status_.COMPLETED)) {

      resultRecord.event = 'click';

      var responseField = this.getResponseFieldType(
        itemfield.fieldtype_maxvalue, itemfield.metertype_maxvalue, undefined, resultRecord);

      response = itemfield._outerinspfieldresult;

      if (response) {
        if (response[responseField] !== undefined) {
          respVal = response[responseField];
        }
      }

      if ($j(optionSelectionFeedbackContainer).find('maximo-checkbox').length > 0) {
        $j(optionSelectionFeedbackContainer).find('maximo-checkbox')[0].checked = respVal;
        if (resultRecord.status_maxvalue === status_.COMPLETED) {
          $j(optionSelectionFeedbackContainer).find('maximo-checkbox')[0].readonly = true;
        }
      }
    }
    this.$.responseContainer.appendChild(wrapper);
    this.$.responseContainer.appendChild(optionSelectionFeedbackContainer);

    // Apply style
    label.setAttribute('style', 'display: block;');
    label.$.label.setAttribute('style', 'font-size:1.286em; color:' + this.getComputedStyleValue('--Primary-gray50') + ';');

    optionSelectionFeedbackContainer.setAttribute('style', 'display: flex; position: relative;');

    wrapper.setAttribute('style', 'display: flex; position: relative;');
    wrapper.querySelectorAll('input').forEach(function(input) {
      input.setAttribute('style', 'margin-left: 15px;margin-right: 5px;display:none;');
    });
  },

  _setOptionStyle: function(optionWrapper, radioBtn, radioLabel, wrapper1, wrapper) {

    radioLabel.appendChild(radioBtn);
    optionWrapper = document.createElement('div');
    optionWrapper.appendChild(radioLabel);
    radioLabel.setAttribute('style', 'border: 1px solid ' + this.getComputedStyleValue('--Secondary-blue20') + '; padding: 5px; margin: 8px 0px 8px 0px;');
    radioLabel.style.display = 'flex';
    radioLabel.style.flexDirection = 'row';
    radioLabel.style.flexWrap = 'nowrap';
    radioLabel.style.alignItems = 'center';

    wrapper1.appendChild(optionWrapper);
    wrapper1.setAttribute('style', 'width: 48%; margin-right: 2%;');
    wrapper.appendChild(wrapper1);

    return wrapper;
  },

  
  
  /* 
   * Function to format the Single Option items, according to the rules:
   * 
   * if there´s an answer, then there´s a selected option - format:
   * 
   * selected option:
   * if colors:   background-color and border in the option color
   * if white/no color:   background is green and border is blue
   * 
   * if there´s no selected option:
   * if colors: background-color white and border in the option color
   * if white/no color: background is white and border is blue
   * 
   * 
   * if there´s no answer, then there´s no selected item - format:
   * all the options:
   * if colors: background-color and border in the option color
   * if white/no color: background is white and the border is blue
   * 
   */
  _setSingleOptionStyle: function(respVal, optionWrapper, optionWrapperCol2, radioBtn, radioLabel, col, wrapper1, wrapper2, wrapper, status) {

    if (radioBtn.icon && !(radioBtn.icon === "")) {
      iconNode = Polymer.Base.create('iron-icon', {'icon': radioBtn.icon});
      iconNode.style.margin = '0 5px';

      radioLabel.insertBefore(iconNode, radioLabel.firstChild);
    }

    radioLabel.appendChild(radioBtn);

    if (col === '1') {
      optionWrapper = document.createElement('div');
      optionWrapper.setAttribute('class', 'optionWrapper');
      optionWrapper.appendChild(radioLabel);

    } else if (col === '2') {
      optionWrapperCol2 = document.createElement('div');
      optionWrapperCol2.setAttribute('class', 'optionWrapper');
      optionWrapperCol2.appendChild(radioLabel);
    }

    radioLabel.setAttribute('class', 'divOption');
    radioLabel.setAttribute('style', 'border: 1px solid; padding: 5px; margin: 8px 0px 8px 0px;');
    radioLabel.style.display = 'flex';
    radioLabel.style.flexDirection = 'row';
    radioLabel.style.flexWrap = 'nowrap';
    radioLabel.style.alignItems = 'center';


    if (respVal) { /* there´s an item selected in the list */

      if (radioBtn.checked) { /* item selected */

        if (radioBtn.color && !(radioBtn.color === "") && !(radioBtn.color.toUpperCase() === this.getComputedStyleValue('--Primary-white').toUpperCase())) {
          radioLabel.style.backgroundColor = radioBtn.color;
          radioLabel.style.borderColor = radioBtn.color;
        } else {
          radioLabel.style.backgroundColor = this.getComputedStyleValue('--Primary-green10');
          radioLabel.style.borderColor = this.getComputedStyleValue('--Secondary-blue20');
        }
      } else { /* non selected items */
        if (status === status_.COMPLETED) {
          radioLabel.style.backgroundColor = this.getComputedStyleValue('--Neutral-gray4');
          radioLabel.style.borderColor = this.getComputedStyleValue('--Neutral-gray4');
        } else {
          if (radioBtn.color && !(radioBtn.color === "") && !(radioBtn.color.toUpperCase() === this.getComputedStyleValue('--Primary-white').toUpperCase())) {
            radioLabel.style.backgroundColor = this.getComputedStyleValue('--Primary-white')
            radioLabel.style.borderColor = radioBtn.color;
          } else {
            radioLabel.style.backgroundColor = this.getComputedStyleValue('--Primary-white')
            radioLabel.style.borderColor = this.getComputedStyleValue('--Secondary-blue20');
          }
        }
      }

    } else { /* there´s no item selected in the list */

      if (status === status_.COMPLETED) {
        radioLabel.style.backgroundColor = this.getComputedStyleValue('--Neutral-gray4');
        radioLabel.style.borderColor = this.getComputedStyleValue('--Neutral-gray4');
      } else {
        if (radioBtn.color && !(radioBtn.color === "") && !(radioBtn.color.toUpperCase() === this.getComputedStyleValue('--Primary-white').toUpperCase())) {
          radioLabel.style.backgroundColor = radioBtn.color;
          radioLabel.style.borderColor = radioBtn.color;
        } else {
          radioLabel.style.backgroundColor = this.getComputedStyleValue('--Primary-white');
          radioLabel.style.borderColor = this.getComputedStyleValue('--Secondary-blue20');
        }
      }

    }

    if (col === '1') {
      wrapper1.appendChild(optionWrapper);
      wrapper1.setAttribute('style', 'width: 48%; margin-right: 2%;');
      wrapper.appendChild(wrapper1);
    } else if (col === '2') {
      wrapper2.appendChild(optionWrapperCol2);
      wrapper2.setAttribute('style', 'width: 48%;');
      wrapper.appendChild(wrapper2);
    }
    return wrapper;

  },


  /**
   * Set Inspection Message and Confirm Box
   */
  _setInspFeedback: function(OptionMsgDiv, itemfield, OptionConfirmDiv, OptionMainDiv, wrapper, option, self, resultRecord) {

    var checkbox = '';
    $j('#OptionMsgDiv_' + itemfield._id).remove();
    $j('#OptionConfirmDiv_' + itemfield._id).remove();
    OptionMainDiv.hidden = true;
    if (option.inspectorfeedback !== undefined && option.inspectorfeedback !== '') {
      OptionMsgDiv = Polymer.Base.create('div', {
        'id': 'OptionMsgDiv_' + itemfield._id
      });

      OptionMsgDiv.appendChild(Polymer.Base.create('iron-icon', {
        'id': 'Optionmessageicon_' + itemfield._id,
        'icon': 'Maximo:Warning'
      }));

      OptionMsgDiv.appendChild(Polymer.Base.create('maximo-label', {
        'id': 'Optionmessagelabel_' + itemfield._id,
        'label': option.inspectorfeedback
      }));

      OptionMsgDiv.setAttribute('style', 'padding: 5px 0px 0px 5px; display: flex;');
      OptionMsgDiv.firstElementChild.setAttribute('style', 'color: #fde876;');
      OptionMainDiv.appendChild(OptionMsgDiv);
      OptionMainDiv.hidden = false;
    }

    if (itemfield.readconfirmation && option.inspectorfeedback !== undefined && option.inspectorfeedback !== '') {

      OptionConfirmDiv = Polymer.Base.create('div', {
        'id': 'OptionConfirmDiv_' + itemfield._id
      });

      OptionConfirmDiv.appendChild(Polymer.Base.create('maximo-label', {
        'id': 'Optioncheckboxlabel_' + itemfield._id,
        'label': $M.localize('uitext', 'mxapiinspection', 'confirm')
      }));

      OptionConfirmDiv.setAttribute('style', 'padding-left: 15px; white-space: nowrap;');

      checkbox = document.createElement('maximo-checkbox');
      checkbox.id = 'Optionmessagecheckbox_' + itemfield._id;
      checkbox.checked = false;

      OptionConfirmDiv.appendChild(checkbox);
      OptionMainDiv.appendChild(OptionConfirmDiv);
      OptionMainDiv.hidden = false;

      // Set proper ID
      if ('_setIdentification' in checkbox) {
        checkbox._setIdentification();
      }

      checkbox.addEventListener('click', function(event) {
        var response = itemfield._outerinspfieldresult;
        var lastResponse = self.fieldResponse ? self.fieldResponse : response;
        self._checkBoxEventListener(event, self, resultRecord, itemfield, lastResponse, this.checked, self, resultRecord);
      });
    }

    wrapper.appendChild(OptionMainDiv);
    return wrapper;
  },

  /**
   * Function call when click on checkbox
   */
  _checkBoxEventListener: function(event, self, resultRecord, itemfield, lastResponse, checked) {
    event.stopPropagation();
    self.clearMessages();

    resultRecord.event = event.type;
    var payload = {
      inspresult: resultRecord,
      respfield: self.prepareResponseField(resultRecord, itemfield, checked, lastResponse),
      successCallback: self.updateResponseSuccessHandler.bind(self),
      errorCallback: self.updateResponseFailureHandler.bind(self)
    };

    self.fire('update-response-field', payload);
  },

  /**** Attachhments section *****/

  /**
   * Build attachment response field
   */
  buildAttachmentsInput: function() {

    var resultRecord = this.resultrecord;
    var itemfield = this.item;

    // Create component attributes
    var labelAttrs = {};
    labelAttrs.id = itemfield._id + '_attachmentLabel';
    labelAttrs.label = itemfield.description;

    var fileUpAttrs = {};
    fileUpAttrs.id = itemfield._id + '_attachmentFileUpload';

    var fileCarouselAttrs = {};
    fileCarouselAttrs.id = itemfield._id + '_attachmentFileCarousel';
    
    var responseFieldType = this.getResponseFieldType(
        itemfield.fieldtype_maxvalue, itemfield.metertype_maxvalue);

    // Set require flag
    if (itemfield.required === true) {
      this._setRequired(true);
      fileUpAttrs.required = true;
    } else {
      this._setRequired(false);
    }

    // Create components
    var label = Polymer.Base.create('maximo-label', labelAttrs);
    var fileUpload = Polymer.Base.create('maximo-file-upload', fileUpAttrs);
    var fileCarousel = Polymer.Base.create('maximo-file-carousel', fileCarouselAttrs);
    
    
    // Set readonly or disabled
    if (!resultRecord || resultRecord.status_maxvalue === 'COMPLETED') {
      fileUpload.disable();
      fileCarousel.disableDeletion(true);
    }

    fileCarousel.enableButtons();
    
    // Fetching response 
    var response = itemfield._outerinspfieldresult;

    // Set listener for upload completion
    this.listen(fileUpload, 'uploadComplete', 'reloadInspectionResult');

    // Prevent scroll to top
    label.$.label.addEventListener('click', this.preventScroll);
    
    // Insert new component
    this.$.responseContainer.appendChild(label);
    this.$.responseContainer.appendChild(fileUpload);
    this.$.responseContainer.appendChild(fileCarousel);

    // Set proper ID
    if ('_setIdentification' in fileUpload) {
      fileUpload._setIdentification();
    }
    if ('_setIdentification' in fileCarousel) {
      fileCarousel._setIdentification();
    }

    fileUpload.fileDocType = itemfield.doctype;

    // Set value and store it locally
    if (response) {
      
      fileUpload.memberData = response;
      // Unecessary since completion has been processed before in container
      this.async(function() {
        this._responseChanged(responseFieldType, response);
      });
      let carouselAttachments = [];
      if (response.doclinks && 
        response.doclinks.member &&
        response.doclinks.member.length) {

        carouselAttachments = this.buildCarouselAttachments(response.doclinks.member);
      }
      this.updateAttachments(carouselAttachments, response);

    } else if (!resultRecord) {
      // It is preview mode
      // Do nothing
    } else {

      this.updateAttachments([]);
      /**
       * Make sure to have fieldresult
       * so it can fetches its doclink href
       */
      var payload = {
        inspresult: resultRecord,
        respfield: this.prepareResponseField(resultRecord, itemfield, null, response),
        successCallback: this.updateResponseSuccessHandler.bind(this),
        errorCallback: this.updateResponseFailureHandler.bind(this)
      };

      this.async(function() {
        this.fire('update-response-field', payload);
      });
    }

    // Add style
    label.$.label.setAttribute('style', 'font-size:1.286em; color: #777677;');
    fileCarousel.customStyle['--img-size'] = '40%';

    fileCarousel.updateStyles();
    fileUpload.updateStyles();

  },

  /**
   * Build unavailable response field
   */
   buildUnavailableInputPreview: function(questionType) {

    var itemfield = this.item;

    // Create component attributes for the label
    var labelAttrs = {};
    labelAttrs.id = itemfield._id + '_unavailableLabel';
    labelAttrs.label = itemfield.description;

    // Create components for the label
    var label = Polymer.Base.create('maximo-label', labelAttrs);
    // Insert new component
    this.$.responseContainer.appendChild(label);
    // Add style
    label.$.label.setAttribute('style', 'font-size:1.286em; color: #777677;');

    // Create component attributes for the message
    var msgAttrs = {};
    msgAttrs.id = itemfield._id + '_unavailableMessage';
    msgAttrs.label = $M.localize('uitext', 'mxapiinspresult','feature_unavailable',[questionType]);

    // Create component for the message
    var msgNode = Polymer.Base.create('div', {'id': msgAttrs.id });

    msgNode.appendChild(Polymer.Base.create('iron-icon', {
        'id': 'SigMessageicon_' + msgAttrs.id,
        'icon': 'Maximo:Warning'
      }));

    var msgUnavailable= Polymer.Base.create('maximo-label', msgAttrs);
    msgNode.appendChild(msgUnavailable);
    
    //Append node to container
    this.$.msgContainer.appendChild(msgNode);
  },

  /**
   * Pick only useful informations of each doclink
   * @return lapidatedArray
   */
  buildCarouselAttachments: function(attachmentMembers) {
    var lapidatedArray = [];
    for (var i = 0; i < attachmentMembers.length; i++) {
      var item = attachmentMembers[i];
      var imgName = item.describedBy.description;
      var imgPath = item.href;
      var imageObj = [ imgName, imgPath ];
      lapidatedArray.push(imageObj);
    }
    return lapidatedArray;
  },

  /**
   * Listener confirms attachment deletion
   * from mx-file-carousel
   */
  _deleteAttachment: function(event) {

    var resultRecord = this.resultrecord;
    var attachment = event.detail[1];

    // Request item deletion
    var payload = {
        inspresult: resultRecord,
        attachment: attachment,
        successCallback: this.deleteAttachmentSuccessHandler.bind(this),
        errorCallback: this.deleteAttachmentFailureHandler.bind(this)
    };

    this.fire('delete-attachment', payload);
  },
  
  /**
   * Listener confirms attachment upload
   * from mx-file-upload
   */
  reloadInspectionResult: function(event) {
    var resultRecord = this.resultrecord;

    // Fetch doclinks when upload is complete
    var payload = {
        inspresult: resultRecord,
        successCallback: this.fetchResultSuccessHandler.bind(this),
        errorCallback: this.fetchResultFailureHandler.bind(this)
    };

    this.fire('fetch-inspection-result', payload);
  },
  
  /**
   * Receive new attachment set prepare presentation
   */
  updateAttachments: function(attachmentsArray, response) {

    //this.attachments = attachmentsArray;
    var selectedImage;
    if (attachmentsArray) {
      // Try finding any file carousel before creating one
      var fileCarousel = this.$.responseContainer.querySelector('maximo-file-carousel');
      selectedImage = fileCarousel.selected;

      fileCarousel.hidden = attachmentsArray.length <= 0;

      if (selectedImage && attachmentsArray.length > 0 && 
          selectedImage > (attachmentsArray.length - 1)) {
        selectedImage = attachmentsArray.length - 1;
      }

      fileCarousel.images = attachmentsArray;
      fileCarousel.selected = selectedImage ? selectedImage : 0;
    }
    
    if (response) {
      var fileUpload = this.$.responseContainer.querySelector('maximo-file-upload');
      var fileCarousel = this.$.responseContainer.querySelector('maximo-file-carousel');
      fileUpload.memberData = response;
      fileCarousel.enableButtons();
    }
  },

  

  /****** Meter Input ******/
  
  /**
   * Build input component
   */
  buildMeterInput: function() {

    var itemfield = this.item;
    switch (itemfield.metertype_maxvalue) {
      case 'CHARACTERISTIC':
        this.isCharacteristicMeter = true;
        // Build characteristic meter only when domain fetch finishes
        this.async(function(){
          this.fetchCharacteristicMeterOptions(itemfield.metername).then(function(response){
            // Grab domain values
            var values = this.extractDomainValues(response.response.member);
            this.clearContainers();
            this.buildCharacteristicMeterInput(values);
          }.bind(this)).catch(function (reason) {
            console.error('An error occurred', reason);
          });
        });
        break;
      case 'CONTINUOUS':
        this.async(function(){
          this.buildContinuousMeterInput();
        });
        break;
      case 'GAUGE':
        this.async(function(){
          this.buildGaugeMeterInput();
        });  
        break;
      default :
        this.async(function(){
          this.buildGaugeMeterInput();
        });
        break;
    }
  },

  /**
   * Build Gauge Meter
   */
  buildGaugeMeterInput: function() {
    // Set base vars
    var self = this;
    var resultRecord = this.resultrecord;
    var itemfield = this.item;
    // Fetching response 
    var response = itemfield._outerinspfieldresult;

    // Get field type
    var responseFieldType = this.getResponseFieldType(
        itemfield.fieldtype_maxvalue, itemfield.metertype_maxvalue);
    
    // Create component attributes
    var inputAttrs = {};
    inputAttrs.id = itemfield._id + '_metertextresponse';
    inputAttrs.type = 'number';
    inputAttrs.label = itemfield.description;
    if (resultRecord && resultRecord.status_maxvalue !== 'COMPLETED') {
      inputAttrs.placeholder = $M.localize('uitext', 'mxapiinspresult','enter_meter_reading');
    }

    var metersFound = this.findInspectionMeters(this.meters, itemfield.metername);

    if (!metersFound.length) {
      var msg = $M.localize('uitext', 'mxapiinspresult', 'meter_unavailable',[ itemfield.metername ]);
      // Include notification
      this.showError(msg);
    }

    // Set readonly
    if (!resultRecord || !metersFound.length 
      || (resultRecord && resultRecord.status_maxvalue === 'COMPLETED')
      || (response && response.errorflag === 2)) {
    	  inputAttrs.readonly = true;
    }
    
    // Set require flag
    if (itemfield.required === true) {
      if (metersFound.length || !resultRecord) {
        this._setRequired(true);
      } else {
        this._setRequired(false);
        this._invalidateResponse();
      }
    }

    // Create component
    var responseField = Polymer.Base.create('maximo-text', inputAttrs);
    
    // Add attribute step = 'any' into the input inside maximo-text. Defect 310674
    var inputNode = responseField.querySelector('input');
    inputNode.setAttribute('step','any');
    
    // Add change listener on input to restore last value case invalid input
    responseField.addEventListener('input', function(event) {
      if (!this.$.input.checkValidity()) {
        self.restorePreviousValue();
      }
    });

    // Add change listener on input
    responseField.addEventListener('change', function(event) {
      self.clearMessages();

      var val = (this.value) ? Number(this.value) : this.value;
      var lastResponse = self.fieldResponse ? self.fieldResponse : response;
      
      var payload = {
          inspresult: resultRecord,
          respfield: self.prepareResponseField(resultRecord, itemfield, val, lastResponse),
          successCallback: self.updateResponseSuccessHandler.bind(self),
          errorCallback: self.updateResponseFailureHandler.bind(self)
      };

      self.fire('update-response-field', payload);
    });
    
    // Prevent scroll to top
    responseField.$.label.addEventListener('click', this.preventScroll);
    
    // Append input in document
    this.$.responseContainer.appendChild(responseField);
    // Set proper ID
    if ('_setIdentification' in responseField) {
      responseField._setIdentification();
    }

    // Need to be after inserted in DOM tree
    // Set value and store it locally
    if (response) {
      if (response[responseFieldType] !== undefined) {
        var respVal = response[responseFieldType];
        responseField.value = respVal;
        // Unecessary since completion has been processed before in container
        // Store response object
        this.async(function() {
          this._responseChanged(responseFieldType, response);
        });
      }

      if (response.errormessage && response.errormessage.length) {
        this.showError(response.errormessage);
      }

    }
    
    // Add style
    responseField.setAttribute('style', 'text-align:left;');
    responseField.$.label.setAttribute('style', 'font-size:129%;');
    responseField.$.input.setAttribute('style', 'font-size:129%; width:50%;');
    //responseNode.$.input.setAttribute('id', itemfield._id + '_input_response');
  },

  /**
   * Build Continuour Meter
   */
  buildContinuousMeterInput: function () {

    // Set base vars
    var self = this;
    var resultRecord = this.resultrecord;
    var itemfield = this.item;
    // Fetching response 
    var response = itemfield._outerinspfieldresult;

    // Get field type
    var responseFieldType = this.getResponseFieldType(
        itemfield.fieldtype_maxvalue, itemfield.metertype_maxvalue);
    
    // Create component attributes
    var inputAttrs = {};
    inputAttrs.id = itemfield._id + '_metertextresponse';
    inputAttrs.type = 'number';
    inputAttrs.label = itemfield.description;
    if (resultRecord && resultRecord.status_maxvalue !== 'COMPLETED') {
      inputAttrs.placeholder = $M.localize('uitext', 'mxapiinspresult','enter_meter_reading');
    }

    var checkboxAttrs = {};
    checkboxAttrs.id = itemfield._id + 'meterrollover';
    checkboxAttrs.label = $M.localize('uitext', 'mxapiinspresult', 'rollover');
    
    var metersFound = this.findInspectionMeters(this.meters, itemfield.metername);

    if (!metersFound.length) {
      var msg = $M.localize('uitext', 'mxapiinspresult', 'meter_unavailable',[ itemfield.metername ]);
      // Include notification
      this.showError(msg);
    }

    // Set readonly
    if (!resultRecord || !metersFound.length  
      || (resultRecord && resultRecord.status_maxvalue === 'COMPLETED')
      || (response && response.errorflag === 2)) {
      inputAttrs.readonly = true;
      checkboxAttrs.readonly = true;
    }

    // Set require flag
    if (itemfield.required === true) {
      if (metersFound.length || !resultRecord) {
        this._setRequired(true);
      } else {
        this._setRequired(false);
        this._invalidateResponse();
      }
    }

    // Create component
    var responseField = Polymer.Base.create('maximo-text', inputAttrs);
    
    // Add attribute step = 'any' into the input inside maximo-text. Defect 310674
    var inputNode = responseField.querySelector('input');
    inputNode.setAttribute('step','any');
    
    var additionalField = Polymer.Base.create('maximo-checkbox', checkboxAttrs);
    additionalField.dataset.name = 'rollover';
    
    // Add change listener on input to restore last value case invalid input
    responseField.addEventListener('input', function(event) {
      if (!this.$.input.checkValidity()) {
        self.restorePreviousValue();
      }
    });

    // Add change listener on input
    responseField.addEventListener('change', function(event) {
      self.clearMessages();

      var val = (this.value) ? Number(this.value) : this.value;
      var lastResponse = self.fieldResponse ? self.fieldResponse : response;
      
      var payload = {
          inspresult: resultRecord,
          respfield: self.prepareResponseField(resultRecord, itemfield, val, lastResponse),
          successCallback: self.updateResponseSuccessHandler.bind(self),
          errorCallback: self.updateResponseFailureHandler.bind(self)
      };

      self.fire('update-response-field', payload);
    });
    // Add change listener on checkbox
    additionalField.addEventListener('change', function(event) {
      self.clearMessages();
      

      var input = self.$.responseContainer.querySelector('maximo-text');
      var val = (input.value) ? Number(input.value) : input.value;
      var lastResponse = self.fieldResponse ? self.fieldResponse : response;
      
      var payload = {
          inspresult: resultRecord,
          respfield: self.prepareResponseField(resultRecord, itemfield, val, lastResponse),
          successCallback: self.updateResponseSuccessHandler.bind(self),
          errorCallback: self.updateResponseFailureHandler.bind(self)
      };

      self.fire('update-response-field', payload);
    });

    // Prevent scroll to top
    responseField.$.label.addEventListener('click', this.preventScroll);

    // Append input in document
    this.$.responseContainer.appendChild(responseField);

    // Set proper ID
    if ('_setIdentification' in responseField) {
      responseField._setIdentification();
    }
    
    // include checkbox only if meter has rollover property and content
    var meter = metersFound[0] || null;
    if (meter && meter.rollover) {
      this.$.responseContainer.appendChild(additionalField);
      // Set proper ID
      if ('_setIdentification' in additionalField) {
        additionalField._setIdentification();
      }
    }

    // Need to be after inserted in DOM tree
    // Set value and store it locally
    if (response) {

      if (response[responseFieldType] !== undefined) {
        var respVal = response[responseFieldType];
        responseField.value = respVal;
        additionalField.checked = response.rolloverflag || false;
        // Unecessary since completion has been processed before in container
        // Store response object
        this.async(function() {
          this._responseChanged(responseFieldType, response);
        });
      }

      if (response.errormessage && response.errormessage.length) {
        this.showError(response.errormessage);
      }
    }

    // Add style
    this.$.responseContainer.classList.add('row');
    responseField.classList.add('flex');

    responseField.setAttribute('style', 'text-align:left;');
    responseField.$.label.setAttribute('style', 'font-size:129%;');
    responseField.$.input.setAttribute('style', 'font-size:129%; width:50%;');

    additionalField.setAttribute('style', 'margin: 0px 20px;');
    additionalField.$.label.setAttribute('style', 'font-size:110%;');

  },

  /**
   * Build Characteristic Meter
   */
  buildCharacteristicMeterInput: function(domainValues) {
    var self = this;
    var resultRecord = this.resultrecord;
    var itemfield = this.item;
    var OptionMsgDiv = '';
    var OptionConfirmDiv = '';

    var respVal;
    var labelAttrs = {};
    labelAttrs.id = itemfield._id + '_attachmentLabel';
    labelAttrs.label = itemfield.description;
    var responseFieldType = this.getResponseFieldType(
      itemfield.fieldtype_maxvalue, itemfield.metertype_maxvalue);

    var metersFound = this.findInspectionMeters(this.meters, itemfield.metername);

    if (!metersFound.length) {
      var msg = $M.localize('uitext', 'mxapiinspresult', 'meter_unavailable', [itemfield.metername]);
      // Include notification
      this.showError(msg);
    }

    if (itemfield.required === true) {
      if (metersFound.length || !resultRecord) {
        this._setRequired(true);
      } else {
        this._setRequired(false);
        this._invalidateResponse();
      }
    }

    var response = itemfield._outerinspfieldresult;

    if (response) {
      // Store response object
      if (response[responseFieldType] !== undefined) {
        respVal = response[responseFieldType];
        // Unecessary since completion has been processed before in container
        this.async(function() {
          this._responseChanged(responseFieldType, response);
        });
      }

      if (response.errormessage && response.errormessage.length) {
        this.showError(response.errormessage);
      }
    }
    // Create component
    var label = Polymer.Base.create('maximo-label', labelAttrs);
    var wrapper = Polymer.Base.create('div');
    var wrapper1 = Polymer.Base.create('div');
    wrapper1.id = 'wrapper1';

    var wrapper2 = Polymer.Base.create('div');
    wrapper2.id = 'wrapper2';

    var OptionMainDiv = Polymer.Base.create('div', {
      'id': 'OptionMainDiv_' + itemfield._id
    });
    OptionMainDiv.setAttribute('style', 'display: flex; justify-content: space-between; width:100%;  bottom: 0; max-height: 40px; overflow-y: auto;');
    var optionSelectionFeedbackContainer = Polymer.Base.create('div', {name: 'optionFeedbackContainer'});

    // Prevent scroll to top
    label.$.label.addEventListener('click', this.preventScroll);

    // Append to container
    this.$.responseContainer.appendChild(label);
    var counter = 0;
    var responseField = domainValues.join(',');
    var option = responseField.split(',');
    var column1 = Math.ceil(option.length / 2);

    for (var i in option) {
      // Iterate over response field option		
      var radioBtn = document.createElement('input');
      var radioId = i + itemfield._id + 'meterselectresponse';

      radioBtn.name = 'radiogroup' + itemfield._id;
      radioBtn.type = 'radio';
      radioBtn.value = option[i];
      radioBtn.id = radioId;

      if (respVal && respVal === radioBtn.value) {
        radioBtn.checked = true;
      }

      // Disable button case no
      if (!resultRecord || !metersFound.length
        || resultRecord.status_maxvalue === 'COMPLETED'
        || (response && response.errorflag === 2)) {
        radioBtn.disabled = true;
      }

      radioBtn.addEventListener(this.singleOptionEventType, function(event) {

        event.stopPropagation();
        self.clearMessages();
        for (var i = 0; i < wrapper1.childElementCount; i++) {
          $j(wrapper1.children[i]).find('label').css('backgroundColor', self.getComputedStyleValue('--Primary-white'));
        }

        for (i = 0; i < wrapper2.childElementCount; i++) {
          $j(wrapper2.children[i]).find('label').css('backgroundColor', self.getComputedStyleValue('--Primary-white'));
        }

        var lastResponse = self.fieldResponse ? self.fieldResponse : response;

        var tempRadioValue = this.value;

        if ((lastResponse !== undefined) && (tempRadioValue === lastResponse.txtresponse)) {
          tempRadioValue = '';
        } else {
          $j(event.target.parentElement).css('backgroundColor', self.getComputedStyleValue('--Primary-green10'));
        }

        resultRecord.event = "change";
        var payload = {
          inspresult: resultRecord,
          respfield: self.prepareResponseField(resultRecord, itemfield, tempRadioValue, lastResponse),
          successCallback: self.updateResponseSuccessHandler.bind(self),
          errorCallback: self.updateResponseFailureHandler.bind(self)
        };

        self.fire('update-response-field', payload);
      });

      var optionWrapper = document.createElement('div');

      var radioLabel = document.createElement('label');
      radioLabel.setAttribute('for', radioId);
      radioLabel.setAttribute('id', radioId + '_label');

      // Accessibility aspect for single option
      radioLabel.setAttribute('role','radio');
      radioLabel.setAttribute('tabindex', '0');
      radioLabel.addEventListener('keydown', function(event) {
        switch (event.which){
          case 13: // enter
          case 32: // space
            event.preventDefault();
            var radioInput = this.querySelector('input');
            if (radioInput) radioInput.click();
            break;
        }
      });

      radioLabel.textContent = option[i];
      if (option.length <= 4) {
        wrapper = this._setOptionStyle(optionWrapper, radioBtn, radioLabel, wrapper1, wrapper);
      } else {
        if (counter < column1) {
          wrapper = this._setOptionStyle(optionWrapper, radioBtn, radioLabel, wrapper1, wrapper);
        } else {
          radioLabel.appendChild(radioBtn);
          var optionWrapperCol2 = document.createElement('div');
          optionWrapperCol2.appendChild(radioLabel);
          
          radioLabel.setAttribute('style', 'border: 1px solid ' + this.getComputedStyleValue('--Secondary-blue20') + '; padding: 5px; margin: 8px 0px 8px 0px;');
          radioLabel.style.display = 'flex';
          radioLabel.style.flexDirection = 'row';
          radioLabel.style.flexWrap = 'nowrap';
          radioLabel.style.alignItems = 'center';
          
          wrapper2.appendChild(optionWrapperCol2);
          wrapper2.setAttribute('style', 'width: 48%');
          wrapper.appendChild(wrapper2);

        }
      }
      counter++;
      if (radioBtn.checked) {
        if (resultRecord.status_maxvalue !== 'COMPLETED') {
          radioBtn.parentNode.style.color = this.getComputedStyleValue('--Primary-green100');
          radioBtn.parentNode.style.backgroundColor = this.getComputedStyleValue('--Primary-green10');
        }
        else if (resultRecord.status_maxvalue === 'COMPLETED') {
          radioBtn.parentNode.style.backgroundColor = this.getComputedStyleValue('--Neutral-gray4');
        }
        optionSelectionFeedbackContainer = this._setInspFeedback(OptionMsgDiv, itemfield, OptionConfirmDiv, OptionMainDiv, optionSelectionFeedbackContainer, option, self, resultRecord);
      }
    }


    this.$.responseContainer.appendChild(wrapper);
    this.$.responseContainer.appendChild(optionSelectionFeedbackContainer);

    // Apply style
    label.setAttribute('style', 'display: block;');
    label.$.label.setAttribute('style', 'font-size:1.286em; color ' + this.getComputedStyleValue('--Primary-green50') + ';');

    wrapper.setAttribute('style', 'display: flex; position: relative;');
    wrapper.querySelectorAll('input').forEach(function(input) {
      input.setAttribute('style', 'margin-left: 15px;margin-right: 5px; display:none;');
    });

    optionSelectionFeedbackContainer.setAttribute('style', 'display: flex; position: relative; margin-bottom: 10px;');
  },

  /**
   * Fetch meter options 
   */
  fetchCharacteristicMeterOptions: function(meterName) {

    //var collection = this.$.meterOptionsCollection;
    var collection = this.querySelector('maximo-collection');
    var maximoQbe = new MaximoQbe();
    maximoQbe.setEquals('metername', meterName);
    var qbeObject = maximoQbe.toJSON();
    collection.queryWhere = qbeObject;

    return collection.refreshRecords();
  },

  /**
   * Prepare response to load as value
   */
  _handleMeterDataResponse: function(e) {
    var collection = this.$.meterOptionsCollection;
    var domainArray = collection.collectionData;
    var values = [];
    if (domainArray.length > 0) {
      var dmValues = domainArray[0].domainid;
      if (dmValues.length > 0) {
        var alnValues = dmValues[0].alndomainvalue;
        values = alnValues.map(function(synonym) {
          return synonym.value;
        });
      }
    }

    this.loadCharacteristicMeterOptions(values);
  },

  /**
   * Prepare empty response to load as value
   */
  _handleMeterDataError: function(e) {

    var values = [];
    this.loadCharacteristicMeterOptions(values);
  },

  /**
   * Load options for Characteristic meters
   */
  loadCharacteristicMeterOptions: function(valueArray) {

    var node = this.querySelector('maximo-select');
    node.values = valueArray.join(',');

  },

  /**** Callback handlers ****/

  /**
   * Success create/update request handler
   * Persist data callback from mx-execution
   */
  updateResponseSuccessHandler: function(response){

    // Check value
    var inspFieldResults = response.inspfieldresult;
    var itemField = this.item;

    // Find inspfieldresult in array
    var fieldResult = this.findFieldResultByItem(inspFieldResults, itemField);

    if (fieldResult) {
      //Store new value to response property
      var responseFieldType = this.getResponseFieldType(
          itemField.fieldtype_maxvalue, itemField.metertype_maxvalue,response.datetimeField);

      this.async(function() {
        let fireEvent = true;
        this._responseChanged(responseFieldType, fieldResult, fireEvent);
      });
    }

    // Set new inspfieldresult into inspresult
    this.resultrecord.inspfieldresult = inspFieldResults;

  },
  
  /**
   * Fail create/update request handler
   * Persist data callback from mx-execution
   */
  updateResponseFailureHandler: function(error){
    this.restorePreviousValue();
    //Show error in msgContainer div
    this.showError(error);
  },

  /**
   * Success load insp result handler
   * Load record callback from mx-execution
   */
  fetchResultSuccessHandler: function(response) {

    // Check value
    var inspFieldResults = response.inspfieldresult;
    var itemField = this.item;

    // Find inspfieldresult in array
    var fieldResult = this.findFieldResultByItem(inspFieldResults, itemField);

    if (fieldResult) {

      //Store new value to response property
      var responseFieldType = this.getResponseFieldType(
          itemField.fieldtype_maxvalue, itemField.metertype_maxvalue);

      this.async(function() {
        let fireEvent = true;
        this._responseChanged(responseFieldType, fieldResult, fireEvent);
      });

    }

    // Set new inspfieldresult into inspresult
    this.resultrecord.inspfieldresult = inspFieldResults;
  },
  
  /**
   * Fail load insp result handler
   * Load record callback from mx-execution
   */
  fetchResultFailureHandler: function(event) {
    console.error('Problem to load inspection result');
  },
  
  /**
   * Fail deleting attachment handler
   * Delete attachment error callback from mx-execution
   */
  deleteAttachmentSuccessHandler: function(event) {
    // Fetch field result doclinks
    this.reloadInspectionResult(event);
  },
  
  /**
   * Fail deleting attachment handler
   * Delete attachment error callback from mx-execution
   */
  deleteAttachmentFailureHandler: function(event) {
    console.error('Problem to delete attachment');
  },

  /****** Utils ******/

  /**
   * Reflects property to field
   */
  itemPropertySetter: function(property, value) {
	  if (this.item && this.item[property] !== value) {
		  this.item[property] = value;
	  } 
  },
  
  /**
   * Prepare new responsefield to be created/updated
   * @return inspfieldresult
   */
  prepareResponseField: function(resultRec, fieldRec, newValue, existingResponse,dateTimeField) {
    var inspFieldRes = {};

    inspFieldRes.resultnum = resultRec.resultnum;
    inspFieldRes.orgid = resultRec.orgid;
    inspFieldRes.siteid = resultRec.siteid;

    inspFieldRes.inspquestionnum = fieldRec.inspquestionnum;
    inspFieldRes.inspformnum = fieldRec.inspformnum;
    inspFieldRes.inspfieldnum = fieldRec.inspfieldnum;
    inspFieldRes.revision = parseInt(fieldRec.revision);
    inspFieldRes.enteredby = $M.userInfo.personid;
    inspFieldRes.entereddate = new Date();

    const responseReference = existingResponse || 
		( resultRec.inspfieldresult ? this.findFieldResultByItem(resultRec.inspfieldresult, fieldRec) : null );

    if (responseReference) {
    	inspFieldRes.href = responseReference.href;
    	inspFieldRes._action = 'Update';
    } else {
    	inspFieldRes._action = 'Add';
    }

    inspFieldRes.metername = (fieldRec.metername) ? fieldRec.metername : '';
    inspFieldRes.rolloverflag = this.fetchRollOverFlagValue();
    
    var responsefield = this.getResponseFieldType(
        fieldRec.fieldtype_maxvalue, fieldRec.metertype_maxvalue,dateTimeField,resultRec);
    
    if(fieldRec.fieldtype_maxvalue==='SO' && responsefield.indexOf('~')>-1){
    
     var responseFieldArray=  responsefield.split('~');
     for(var j=0;j<responseFieldArray.length;j++){
    	 responsefield=responseFieldArray[j];
    	 if(newValue.split('~')[j]==='true'|| newValue.split('~')[j]==='false'){
    		 inspFieldRes[responsefield]=JSON.parse(newValue.split('~')[j]);
    	 }
    	 else{
    	 inspFieldRes[responsefield] = newValue.split('~')[j];
    	 }
     }
    }
    else{
    inspFieldRes[responsefield] = newValue;
    }
    return inspFieldRes;
  },

  /**
   * Restore last valid response to input field
   */
  restorePreviousValue: function() {

    var decimalSeparator = this._getDecimalSeparator();
    var type = this.item.fieldtype_maxvalue;
    var responsefield = this.getResponseFieldType(
        this.item.fieldtype_maxvalue, this.item.metertype_maxvalue);
    var input;
    if (type === 'TR') {
      input = Polymer.dom(this.$.responseContainer).querySelector('maximo-text');
      input.value = this.fieldResponse[responsefield];
    } else if (type === 'SE') {
      input = Polymer.dom(this.$.responseContainer).querySelector('maximo-text');

      if(decimalSeparator === ',') {
        input.value = !this.fieldResponse ? '' : !this.fieldResponse[responsefield + '_localized'] ? '' : this.fieldResponse[responsefield + '_localized'];
      } else {
        input.value = !this.fieldResponse ? '' : !this.fieldResponse[responsefield] ? '' : this.fieldResponse[responsefield];
      }

    } else if (type === 'FU') {
      console.warn('Problem to fetch inspection result data');
    } else if (type === 'MM') {

      var meterType = this.item.metertype_maxvalue;
      if (meterType === 'CHARACTERISTIC') {
        input = Polymer.dom(this.$.responseContainer).querySelector('maximo-select');
      } else {
        input = Polymer.dom(this.$.responseContainer).querySelector('maximo-text');
      }

      if(this.fieldResponse) {
      	input.value = this.fieldResponse[responsefield];
      }
    }

  },

  /**
   * Fetch response from array
   */
  filterResult: function(resultRecord, itemField) {
    var responseArray = [];
    var response = null;
    if (resultRecord && resultRecord.inspfieldresult && 
        itemField.inspfieldresult) {
      responseArray = itemField.inspfieldresult.filter(function(el) {
        return el.orgid === resultRecord.orgid && 
          el.resultnum === resultRecord.resultnum;
      });

      if (responseArray.length > 0 && responseArray[0]) {
        response = responseArray[0];
      }
    }

    return response;
  },

  /**
   * Find the correct fieldresult based on current inspfield (this.item)
   * @return {Object}
   */
  findFieldResultByItem(fieldResultSet, inspField) {
    var response;
    
    var responseArray = fieldResultSet.filter(function(fieldResult) {
      return fieldResult.orgid === inspField.orgid &&  
        fieldResult.inspfieldnum === inspField.inspfieldnum &&  
        fieldResult.inspquestionnum === inspField.inspquestionnum &&
        fieldResult.revision === inspField.revision &&
        fieldResult.inspformnum === inspField.inspformnum;
    });
    
    if (responseArray.length) {
      response = responseArray[0];
      response.fieldtype_maxvalue=inspField.fieldtype_maxvalue;
      if (responseArray.length > 1) {
        console.warn('Multiple matches finding result field');
      }
    }
    
    return response;
  },

  /**
   * Find meters matching current inspection field
   * @return {Array}
   */
  findInspectionMeters: function(availableMeters, fieldMeter) {
	  
	  var meters = availableMeters || [];

    return meters.filter(function(meter) {
      return meter.metername === fieldMeter;
    });

  },

  /**
   * Fetch roll over flag value
   */
  fetchRollOverFlagValue: function() {
    var result = false;
    var checkbox = this.querySelector('maximo-checkbox[data-name="rollover"]');
    if (checkbox) {
      result = checkbox.checked;
    }
    return result;
  },

  /**
   * response change fired by each component
   * Update response field status
   * Fire notification to host
   */
  _responseChanged: function(respType, response, fireEvent) {

    var complete = false;
    if (respType === 'numresponse' && response[respType] || response[respType] === 0) {
      complete = true;
    } else if (respType === 'document' && response.doclinks) {

      var carouselAttachments = [];

      if (response.doclinks.member) {
        //Update attachments
        var attachments = response.doclinks.member;
        if (attachments.length) {
          carouselAttachments = this.buildCarouselAttachments(attachments);
        }
      }

      this.updateAttachments(carouselAttachments, response);
      complete = Boolean(carouselAttachments.length);
      
    }else if(response.fieldtype_maxvalue==='DT'){
        if(response.dateresponse!==undefined && response.timeresponse!==undefined){
            complete = true;
        }     
    }
    else if (response[respType] && response[respType].length) {
      complete = true;
    }

    // Update last response so it can be restored in failure case
    this._setFieldResponse(response);

    if (this.answered === undefined || this.answered !== complete) {
      this._setAnswered(complete);
      if (fireEvent) {
        this.fire('status-changed');
      }
    }
    
  },

  /**
   * depending on component condition
   * response might become invalid
   * Update response field status to complete
   * Fire notification to host
   */
  _invalidateResponse: function() {

    var invalid = true;

    if (this.invalid === undefined || this.invalid !== invalid) {
      this._setInvalid(invalid);
      this.fire('invalid-changed');
    }

  },

  /**
   * Extract values from domain response
   * @return {Array}
   */
  extractDomainValues: function(domainMembers) {
    var values = [];
    if (domainMembers.length > 0) {
      var dmValues = domainMembers[0].domainid;
      if (dmValues.length > 0) {
        var alnValues = dmValues[0].alndomainvalue;
        values = alnValues.map(function(synonym) {
          return synonym.value;
        });
      }
    }
    return values;
  },

  /**
   * Set variables to initial state
   */
  resetComponent: function() {

    // Set response undefined
    this._setFieldResponse(null);
    this._setAnswered(false);
    this._setInvalid(false);

  },

  /**
   * Constructs a div with error message and icon
   */
  buildErrorMessage: function(itemid, msg) {
    //var nodeID = 'errormessage_' + itemid;
    var errorMsgNode = Polymer.Base.create('div', {
      'id': 'errormessage_' + itemid
    });
    errorMsgNode.appendChild(Polymer.Base.create('iron-icon', {
      'id': 'errormessageicon_' + itemid,
      'icon': 'Maximo:Warning'
    }));
    errorMsgNode.appendChild(Polymer.Base.create('maximo-label', {
      'id': 'errormessagelabel_' + itemid,
      'label': msg
    }));
    return errorMsgNode;
  },
  
  /**
   * Get rid everything inside response and msg containers
   */
  clearContainers: function() {
   
    // Clear response field div
    var fieldCont = this.$.responseContainer;
    while (fieldCont.firstChild) {
      fieldCont.removeChild(fieldCont.firstChild);
    }
    
    // Clear msg div
    this.clearMessages();
  },

  /**
   * Manage notification section in response field
   * show existing errors
   */
  showError: function(msg) {
    
    this.clearMessages();

    // if inspection is completed, do not show message.
    if (this.resultrecord && 
      this.resultrecord.status_maxvalue === 'COMPLETED') {
      return;
    }

    var id = this.item._id;
    //Build error message node
    var node = this.buildErrorMessage(id, msg);

    //Append node to container
    this.$.msgContainer.appendChild(node);
  },

  /**
   * Delete error message from response field
   */
  clearMessages: function() {
    //Exclude all children in msg container
    var msgCont = this.$.msgContainer;
    while (msgCont.firstChild) {
      msgCont.removeChild(msgCont.firstChild);
    }
  },
  
  /**
   * Applied to labels
   * Prevent labels to scroll user to the top of the page
   */
  preventScroll: function(event) {
    event.preventDefault();
  },

});
